# A module that imports all ipywidgets

import ipywidgets as W

NUMERIC_SINGLEVALUE_CONTROL_WIDGETS = (
    W.IntSlider,
    # W.FloatSlider,  # floats not supported
    # W.FloatLogSlider,  # floats not supported
    W.BoundedIntText,
    # W.BoundedFloatText,  # floats not supported
    # W.IntText,  # open ended widgets not supported
    # W.FloatText,  # open ended widgets not supported
)
NUMERIC_MULTIVALUE_CONTROL_WIDGETS = (
    W.IntRangeSlider,
    # W.FloatRangeSlider,  # floats not supported
)
NUMERIC_CONTROL_WIDGETS = (
    NUMERIC_SINGLEVALUE_CONTROL_WIDGETS + NUMERIC_MULTIVALUE_CONTROL_WIDGETS
)
NUMERIC_OUTPUT_WIDGETS = NUMERIC_CONTROL_WIDGETS + (
    W.IntProgress,
    W.FloatProgress,
)

BOOLEAN_CONTROL_WIDGETS = (W.ToggleButton, W.Checkbox)
BOOLEAN_OUTPUT_WIDGETS = BOOLEAN_CONTROL_WIDGETS + (W.Valid,)

SELECTION_SINGLEVALUE_CONTROL_WIDGETS = (
    W.Dropdown,
    W.RadioButtons,
    W.Select,
    W.SelectionSlider,
    W.ToggleButtons,
)
SELECTION_MULTIVALUE_CONTROL_WIDGETS = (
    W.SelectionRangeSlider,
    W.SelectMultiple,
)
SELECTION_CONTROL_WIDGETS = (
    SELECTION_SINGLEVALUE_CONTROL_WIDGETS
    + SELECTION_MULTIVALUE_CONTROL_WIDGETS
)
SELECTION_OUTPUT_WIDGETS = SELECTION_CONTROL_WIDGETS

STRING_CONTROL_WIDGETS = (
    # Text,  # open ended widgets not supported
    # Textarea,  # open ended widgets not supported
)
STRING_OUTPUT_WIDGETS = (
    W.Label,
    # W.HTML,  # TODO
    # W.HTMLMath,  # TODO
    # W.Image,  # TODO
)

OTHER_CONTROL_WIDGETS = (
    # W.Button,  # TODO
    # W.Play,  # TODO
    # W.DatePicker,  # TODO
    # W.ColorPicker  # TODO
)

CONTROL_WIDGETS = (
    NUMERIC_CONTROL_WIDGETS
    + BOOLEAN_CONTROL_WIDGETS
    + SELECTION_CONTROL_WIDGETS
    + STRING_CONTROL_WIDGETS
    + OTHER_CONTROL_WIDGETS
)

OUTPUT_WIDGETS = (
    NUMERIC_OUTPUT_WIDGETS
    + BOOLEAN_OUTPUT_WIDGETS
    + SELECTION_OUTPUT_WIDGETS
    + STRING_OUTPUT_WIDGETS
    + (W.Output,)
)

VALUE_WIDGETS = CONTROL_WIDGETS + OUTPUT_WIDGETS
LAYOUT_WIDGETS = (W.Box, W.HBox, W.VBox, W.Accordion, W.Tab)



import copy
import csv
import io
import itertools
import json
import logging

import structlog
from nbconvert.preprocessors import Preprocessor

import illusionist.widgets_str as W
from illusionist import kernel_utils, utils, widgets
from illusionist.client import IllusionistClient
from illusionist.config import settings

log = structlog.get_logger()

structlog.configure(
    wrapper_class=structlog.make_filtering_bound_logger(logging.INFO),
)


WIDGET_ONCHANGE_MIMETYPE = "application/vnd.illusionist.widget-onchange+json"


class IllusionistPreprocessor(Preprocessor, IllusionistClient):
    """
    Executes the notebook and then looks at the widgets available and generates
    a new `onchange` state that is saved into the HTML output.

    This `onchange` state contains all possible values for a widget
    based on the control widget it is affected by.

    ```
    {
        "version_major": 1, "version_minor": 0,
        "all_widgets": [ value_widgets ],
        "control_widgets": [ control_widgets ],
        "onchange": {
            <value_widget_id>: {
                "affected_by": [ control_widget_ids ],
                "values": [ [ ... ] ]
            }
        }
    }
    ```

    This class is based and adapted from nbconvert.preprocessors.execute:
    https://github.com/jupyter/nbconvert/blob/main/nbconvert/preprocessors/execute.py
    """

    def __init__(self, **kw):
        nb = kw.get("nb")
        if nb:
            del kw["nb"]
        Preprocessor.__init__(self, nb=nb, **kw)
        IllusionistClient.__init__(self, nb, **kw)

    def preprocess(self, nb, resources=None, km=None):
        # self.log_level = "DEBUG"
        self.nb = nb
        self.km = km
        self.onchange_values = {}

        resources = resources if resources else {}
        resources["illusionist_devmode"] = settings.dev_mode
        if resources["illusionist_devmode"]:
            self.log.warning(
                "Illusionist Dev Mode is ON. "
                "Output HTML or Notebook will contain extra cells at the end"
            )

        try:
            self.reset_execution_trackers()
            self.execute(cleanup_kc=False)

            # self.widget_onchange_state is created on self.after_notebook
            self.nb.metadata.widgets.update(
                {WIDGET_ONCHANGE_MIMETYPE: self.widget_onchange_state}
            )
        finally:
            # Clean up
            self._cleanup_kernel()

        return nb, resources

    def after_notebook(self):
        """
        Overwrites IllusionistClient

        This is called after the regular notebook cells have been executed

        It iterates all the widgets and generate static properties

        Like Next.JS `getStaticPaths`:
        https://nextjs.org/docs/basic-features/data-fetching#getstaticpaths-static-generation
        """
        # Load helper code into the kernel
        _ = self.exec_code(utils.get_source(widgets))
        _ = self.exec_code(utils.get_source(kernel_utils))

        # Save original Notebook and widget state
        # nb_cells_before = copy.deepcopy(self.nb.cells)
        base_widget_state = copy.deepcopy(self.widget_state)

        value_widgets, control_widgets = self.get_target_widgets()

        # 1. Get a list of how value widgets are affected by control widgets
        affected_by = self.get_affected_widgets(control_widgets, value_widgets)
        # log.msg("Affected by dict", value=affected_by)

        # 2. Iterate affected_by and add matrix (per output widget) to the matrix
        static_values = self.get_static_values(affected_by)
        # log.msg("Static values dict", value=static_values)

        # Save the onchange state
        onChangeState = {"version_major": 1, "version_minor": 0}
        onChangeState["all_widgets"] = value_widgets
        onChangeState["control_widgets"] = control_widgets
        onChangeState["onchange"] = static_values
        self.widget_onchange_state = onChangeState

        # Set the original widget_state back to the original ones
        self.widget_state = base_widget_state

    def get_target_widgets(self):
        value_widgets = self.exec_code("get_widgets_ids(kind='value')")
        value_widgets = self.eval_cell(value_widgets)

        for w in value_widgets:
            name = self.widget_state[w]["_model_name"]
            # log.msg("Value Widget", id=w, model=name)

        control_widgets = self.exec_code("get_widgets_ids(kind='control')")
        control_widgets = self.eval_cell(control_widgets)

        for w in control_widgets:
            name = self.widget_state[w]["_model_name"]
            # log.msg("Control Widget", id=w, model=name)

        return value_widgets, control_widgets

    def get_affected_widgets(self, control_widgets, value_widgets):
        """
        Iterates the control widgets and see which value widgets are affected
        by changes to each of them.

        Returns
        -------
            dictionary of `{value_w_id: <set of control_w_ids that produce a change>}
        """
        affected_by = {m_id: set() for m_id in value_widgets}

        for widget_id in control_widgets:
            init_state = copy.deepcopy(self.widget_state)
            possible_values = self.possible_values(widget_id)
            widget_affects = []

            # Iterate all possible values of the control widget
            # and record if any widget state changed
            for value in possible_values:
                self.exec_code(f"set_widget_value('{widget_id}', {value})")
                new_state = self.widget_state
                diff = diff_state(init_state, new_state, my_id=widget_id)
                widget_affects.extend(diff)

            # For the affected widgets save them
            for affected in widget_affects:
                # The or operation of set will add them to the list
                affected_by[affected] |= {widget_id}

        return affected_by

    def get_static_values(self, affected_by):
        """
        For the list of widgets and its affected links

        Iterate affected_by and create a matrix (per output widget)
        Returns
        -------
            dictionary: { value_widget_id: {
                "affected_by": [control_widget_ids],
                "values": [ ... ]
            }}
        """
        static_values = {}

        for value_w_id, control_w_ids in affected_by.items():
            if len(control_w_ids) > 0:
                values = self.get_static_values_by_widget(
                    value_w_id, control_w_ids
                )
                static_values[value_w_id] = {
                    "affected_by": list(control_w_ids),
                    "values": values,
                }
        return static_values

    def get_static_values_by_widget(self, value_w_id, control_w_ids):
        """
        For a value widget return a matrix of:
        ```
            { hash(control_widget_values): <value_widget value>}
        ```
        Returns
        -------
            dictionary
        """
        # 1. Get all possible combinations control widget values that affect
        # this value widget
        control_w_states = [self.widget_state[w_id] for w_id in control_w_ids]
        product = widget_values_product(control_w_states)

        # 2. Iterate the combinations of possible values to create a matrix
        matrix = {}
        for controls_set in product:
            # Update values of control widgets
            for widget_id, value in zip(control_w_ids, controls_set):
                self.exec_code(f"set_widget_value('{widget_id}', {value})")

            # Save the new value of the output widget
            hash_ = self.hash_fn(control_w_ids)
            matrix[hash_] = self.get_widget_value(value_w_id)

        return matrix

    def get_widget_value(self, widget_id):
        """
        Based on the widget model return the current value for a widget id
        This value is the one we serialize on the onChangeState
        """
        widget_state = self.widget_state[widget_id]
        model_name = widget_state["_model_name"]

        if model_name in W.NUMERIC_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name in W.BOOLEAN_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name in W.SELECTION_OUTPUT_WIDGETS:
            if isinstance(widget_state["index"], tuple):
                return list(widget_state["index"])
            return widget_state["index"]
        elif model_name in W.STRING_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name == "OutputModel":
            return widget_state["outputs"]
        else:
            raise Exception(
                f"Output Widget type '{model_name}' not supported."
            )

    def hash_fn(self, widget_ids):
        """Calculate a hash"""
        widget_states = []
        for widget_id in widget_ids:
            w_state = self.widget_state[widget_id]
            widget_states.append(w_state)

        return hash_fn_plain(widget_states)

    def possible_values(self, widget_id):
        """Return a list of all the possible values for a widget"""
        widget_state = self.widget_state[widget_id]
        return possible_values(widget_state)


def widget_values_product(widget_states):
    """
    Make a product of all the possible widget values
    For each control_widget, get all possible values they can have

    The product is helpful when there are more than 1 control widget
    that change one value widget
    """
    all_possible_values = []
    for w_state in widget_states:
        values = possible_values(w_state)
        all_possible_values.append(values)

    return list(itertools.product(*all_possible_values))


def diff_state(initial, new, my_id=None):
    """
    Return a list of widget model_ids that have changed based on two widget
    states
    """
    diff = []
    for (init_id, init_state), (new_id, new_state) in zip(
        initial.items(), new.items()
    ):
        assert init_id == new_id
        if init_state != new_state:
            diff.append(init_id)

    if my_id and my_id in diff:
        diff.remove(my_id)
    return diff


def possible_values(widget_state):
    """
    Return a list of all the possible values for a single widget
    """
    model_name = widget_state["_model_name"]
    if model_name == "IntRangeSliderModel":
        range_ = range(
            widget_state["min"],
            widget_state["max"] + widget_state["step"],
            widget_state["step"],
        )
        ret = itertools.product(range_, range_)
        ret = [[i, j] for i, j in ret if i <= j]
        return ret
    elif model_name in W.NUMERIC_CONTROL_WIDGETS:
        return list(
            range(
                widget_state["min"],
                widget_state["max"] + widget_state["step"],
                widget_state["step"],
            )
        )
    elif model_name in W.BOOLEAN_CONTROL_WIDGETS:
        return [True, False]
    elif model_name == "SelectionRangeSliderModel":
        range_ = range(0, len(widget_state["_options_labels"]))
        ret = itertools.product(range_, range_)
        ret = [[i, j] for i, j in ret if i <= j]
        return ret
    elif model_name == "SelectMultipleModel":
        range_ = range(0, len(widget_state["_options_labels"]))
        return list(powerset(range_))
    elif model_name in W.SELECTION_CONTROL_WIDGETS:
        range_ = range(0, len(widget_state["_options_labels"]))
        return list(range_)
    else:
        raise Exception(f"Widget type '{model_name}' not supported.")


def powerset(iterable):
    """
    Example
    -------
        powerset([1, 2, 3]) -> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


def hash_fn_plain(widget_states):
    """
    Hash all the possible values of a group of widgets.
    """
    values = []
    for w_state in widget_states:
        model_name = w_state["_model_name"]

        if model_name in W.NUMERIC_CONTROL_WIDGETS:
            value = w_state["value"]
        elif model_name in W.BOOLEAN_CONTROL_WIDGETS:
            value = w_state["value"]
        elif model_name in W.SELECTION_CONTROL_WIDGETS:
            value = w_state["index"]
        else:
            raise Exception(f"Cannot use '{model_name}' as input")

        if isinstance(value, str):
            value = json.dumps(value)
        elif isinstance(value, bool):
            value = json.dumps(value)
        elif isinstance(value, (tuple, list)):
            value = list(value)
            # separators are the same but without the spaces
            value = json.dumps(value, separators=(",", ":"))
        values.append(value)

    return dumps(values)


def dumps(values):
    """
    CSV Serialization
    """
    output = io.StringIO()
    writer = csv.writer(
        output,
        skipinitialspace=True,
        delimiter=",",
        quotechar='"',
        quoting=csv.QUOTE_NONNUMERIC,
    )
    writer.writerow(values)
    contents = output.getvalue()
    output.close()
    return contents.strip()



import os

import illusionist
from illusionist.config import settings


def test_import():
    assert illusionist.__version__ is not None
    assert illusionist.__version__ != "0.0.0"
    assert len(illusionist.__version__) > 0


def test_assets_included():
    template = os.path.join(settings.templates_dir, "illusionist")
    assets = os.path.join(template, "assets")

    assert os.path.exists(os.path.join(template, "conf.json"))
    assert os.path.exists(os.path.join(template, "index.html.j2"))
    assert os.path.exists(os.path.join(assets, "illusionist-embed.css"))
    assert os.path.exists(os.path.join(assets, "illusionist-embed.css.map"))
    assert os.path.exists(os.path.join(assets, "illusionist-embed.js"))
    assert os.path.exists(os.path.join(assets, "illusionist-embed.js.map"))



import os


class Settings:
    def __init__(self, dev_mode=None):
        if dev_mode is None:
            _ = os.environ.get("ILLUSIONIST_DEV_MODE", "")
            dev_mode = False if _ in ["", "0"] else True
        self.dev_mode = dev_mode

        this_dir = os.path.dirname(os.path.realpath(__file__))
        self.templates_dir = os.path.join(this_dir, "templates")


settings = Settings()



import os
import os.path

import jinja2
from nbconvert.exporters.html import HTMLExporter

from illusionist.config import settings
from illusionist.preprocessor import IllusionistPreprocessor


@jinja2.utils.pass_context
def include_external_file(ctx, name):
    """Include an encoded base64 image"""
    with open(os.path.abspath(name), "r") as f:
        content = f.read()
    return jinja2.utils.Markup(content)


class IllusionistHTMLExporter(HTMLExporter):
    # Name for the UI menu item under "File -> Download as"
    export_from_notebook = "Illusionist HTML"
    preprocessors = [IllusionistPreprocessor]
    template_file = "illusionist/index.html.j2"
    extra_template_paths = [settings.templates_dir]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.environment.globals["dev_mode"] = settings.dev_mode

    def _init_resources(self, resources):
        resources = super()._init_resources(resources)
        resources["include_external_file"] = include_external_file
        return resources

    # Keeping this just in case
    # def default_filters(self):
    #     for pair in super().default_filters():
    #         yield pair
    #     yield ("echo_filter", self.echo_filter)

    # def echo_filter(self, text):
    #     return "echo_filter: " + text



import ipywidgets as W
import pytest

import illusionist.widgets as IW


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_static_values_byw_num_nolink(preprocessor, class_):
    preprocessor.exec_code(
        f"""
w = W.{class_.__name__}(min=-3, max=0)
lbl = W.Label()
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    values = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )
    assert list(values.keys()) == ["-3", "-2", "-1", "0"]
    assert list(values.values()) == ["", "", "", ""]


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_static_values_byw_num_link(preprocessor, class_):
    preprocessor.exec_code(
        f"""
w = W.{class_.__name__}(min=-3, max=0)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value * 2)

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    values = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )
    assert list(values.keys()) == ["-3", "-2", "-1", "0"]
    assert list(values.values()) == ["-6", "-4", "-2", "0"]


@pytest.mark.parametrize("class_", IW.NUMERIC_MULTIVALUE_CONTROL_WIDGETS)
def test_static_values_byw_num_multivalue_link(preprocessor, class_):
    preprocessor.exec_code(
        f"""
w = W.{class_.__name__}(min=0, max=2)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value[0] + w.value[1])

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    values = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )

    ans = ['"[0,0]"', '"[0,1]"', '"[0,2]"', '"[1,1]"', '"[1,2]"', '"[2,2]"']
    assert list(values.keys()) == ans
    assert list(values.values()) == ["0", "1", "2", "2", "3", "4"]


@pytest.mark.parametrize("class_", IW.SELECTION_SINGLEVALUE_CONTROL_WIDGETS)
def test_static_values_byw_selection_singlevalue(preprocessor, class_):
    preprocessor.exec_code(
        f"""
options = ["aa", "bb", "cc", "dd"]
w = W.{class_.__name__}(options=options)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value)

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    values = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )

    assert list(values.keys()) == ["0", "1", "2", "3"]
    assert list(values.values()) == ["aa", "bb", "cc", "dd"]


@pytest.mark.parametrize("class_", [W.SelectionRangeSlider])
def test_static_values_byw_selection_multivalue_1(preprocessor, class_):
    preprocessor.exec_code(
        f"""
options = ["a", "b", "c", "d"]
w = W.{class_.__name__}(options=options)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value)

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    values = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )

    ans = [
        '"[0,0]"',
        '"[0,1]"',
        '"[0,2]"',
        '"[0,3]"',
        '"[1,1]"',
        '"[1,2]"',
        '"[1,3]"',
        '"[2,2]"',
        '"[2,3]"',
        '"[3,3]"',
    ]
    assert list(values.keys()) == ans
    ans = [
        '()',
        "('a', 'b')",
        "('a', 'c')",
        "('a', 'd')",
        "('b', 'b')",
        "('b', 'c')",
        "('b', 'd')",
        "('c', 'c')",
        "('c', 'd')",
        "('d', 'd')",
    ]
    assert list(values.values()) == ans


@pytest.mark.parametrize("class_", [W.SelectMultiple])
def test_static_values_byw_selection_multivalue_2(preprocessor, class_):
    preprocessor.exec_code(
        f"""
options = ["a", "b", "c"]
w = W.{class_.__name__}(options=options)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value)

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    control_w_ids = [control_w_id]

    matrix = preprocessor.get_static_values_by_widget(
        value_w_id, control_w_ids
    )
    ans = [
        '"[]"',
        '"[0]"',
        '"[1]"',
        '"[2]"',
        '"[0,1]"',
        '"[0,2]"',
        '"[1,2]"',
        '"[0,1,2]"',
    ]
    assert list(matrix.keys()) == ans
    ans = [
        '()',
        "('a',)",
        "('b',)",
        "('c',)",
        "('a', 'b')",
        "('a', 'c')",
        "('b', 'c')",
        "('a', 'b', 'c')",
    ]
    assert list(matrix.values()) == ans



import ipywidgets as W
import pytest

import illusionist.widgets as IW


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_static_values_num_nolink(preprocessor, class_):
    preprocessor.exec_code(
        f"""
w = W.{class_.__name__}(min=-3, max=0)
lbl = W.Label()
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    value_w_ids = [value_w_id]
    control_w_ids = [control_w_id]

    affected_by = preprocessor.get_affected_widgets(control_w_ids, value_w_ids)
    values = preprocessor.get_static_values(affected_by)

    # nothing link between the widgets
    assert len(values) == 0


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_static_values_num_link(preprocessor, class_):
    preprocessor.exec_code(
        f"""
w = W.{class_.__name__}(min=-3, max=0)
lbl = W.Label()

def update(args):
    lbl.value = str(w.value * 2)

update(None)
w.observe(update, "value")
"""
    )
    control_w_id = preprocessor.exec_code("w.model_id")
    control_w_id = preprocessor.eval_cell(control_w_id)
    value_w_id = preprocessor.exec_code("lbl.model_id")
    value_w_id = preprocessor.eval_cell(value_w_id)
    value_w_ids = [value_w_id]
    control_w_ids = [control_w_id]

    affected_by = preprocessor.get_affected_widgets(control_w_ids, value_w_ids)
    values = preprocessor.get_static_values(affected_by)

    assert list(values.keys()) == value_w_ids
    assert list(values.values()) == [
        {
            "affected_by": control_w_ids,
            "values": {"-1": "-2", "-2": "-4", "-3": "-6", "0": "0"},
        }
    ]



import ipywidgets as W
import pytest

import illusionist.widgets as IW
from illusionist.preprocessor import hash_fn_plain


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_hash_fn_plain_num_single_widget(class_):
    w = class_(min=1, max=5)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == "1"

    w.value = 5
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == "5"


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_hash_fn_plain_num_multiple_widgets(class_):
    w1 = class_(min=1, max=5)
    w2 = class_(min=6, max=10)
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == "1,6"

    w1.value = 5
    w2.value = 8
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == "5,8"


@pytest.mark.parametrize("class_", IW.NUMERIC_MULTIVALUE_CONTROL_WIDGETS)
def test_hash_fn_plain_num_multivalue_single(class_):
    w = class_(min=1, max=7, value=(2, 4))
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[2,4]"'

    w.value = (4, 7)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[4,7]"'


@pytest.mark.parametrize("class_", IW.NUMERIC_MULTIVALUE_CONTROL_WIDGETS)
def test_hash_fn_plain_num_multivalue_multiple(class_):
    w1 = class_(min=1, max=7, value=(2, 4))
    w2 = class_(min=1, max=7, value=(1, 2))
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == '"[2,4]","[1,2]"'

    w1.value = (1, 2)
    w2.value = (5, 7)
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == '"[1,2]","[5,7]"'


@pytest.mark.parametrize("class_", IW.BOOLEAN_CONTROL_WIDGETS)
def test_hash_fn_plain_bool_single(class_):
    w = class_()
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"false"'

    w.value = True
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"true"'


@pytest.mark.parametrize("class_", IW.BOOLEAN_CONTROL_WIDGETS)
def test_hash_fn_plain_bool_multiple(class_):
    w1 = class_()
    w2 = class_()
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == '"false","false"'

    w1.value = True
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == '"true","false"'

    w2.value = True
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn_plain(w_states) == '"true","true"'


@pytest.mark.parametrize("class_", IW.SELECTION_SINGLEVALUE_CONTROL_WIDGETS)
def test_hash_fn_plain_selection_singlevalue(class_):
    options = ["a", "b", "c"]
    w = class_(options=options)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == "0"

    w.value = "b"
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == "1"

    w.value = "c"
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == "2"


@pytest.mark.parametrize("class_", [W.SelectionRangeSlider])
def test_hash_fn_plain_selection_multivalue_1(class_):
    options = ["a", "b", "c", "d"]
    w = class_(options=options)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[0,0]"'

    w.value = ["a", "b"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[0,1]"'

    w.value = ["b", "c"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[1,2]"'

    w.value = ["a", "d"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[0,3]"'


@pytest.mark.parametrize("class_", [W.SelectMultiple])
def test_hash_fn_plain_selection_multivalue_2(class_):
    options = ["a", "b"]
    w = class_(options=options)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[]"'

    w.value = ["a"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[0]"'

    w.value = ["b"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[1]"'

    w.value = ["a", "b"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn_plain(w_states) == '"[0,1]"'



# A module that lists all ipywidgets Model names

NUMERIC_SINGLEVALUE_CONTROL_WIDGETS = (
    "IntSliderModel",
    # "FloatSlider",  # floats not supported
    # "FloatLogSlider",  # floats not supported
    "BoundedIntTextModel",
    # "BoundedFloatText",  # floats not supported
    # "IntText",  # open ended widgets not supported
    # "FloatText",  # open ended widgets not supported
)
NUMERIC_MULTIVALUE_CONTROL_WIDGETS = (
    "IntRangeSliderModel",
    # "FloatRangeSlider",  # floats not supported
)
NUMERIC_CONTROL_WIDGETS = (
    NUMERIC_SINGLEVALUE_CONTROL_WIDGETS + NUMERIC_MULTIVALUE_CONTROL_WIDGETS
)
NUMERIC_OUTPUT_WIDGETS = NUMERIC_CONTROL_WIDGETS + (
    "IntProgressModel",
    "FloatProgressModel",
)

BOOLEAN_CONTROL_WIDGETS = ("ToggleButtonModel", "CheckboxModel")
BOOLEAN_OUTPUT_WIDGETS = BOOLEAN_CONTROL_WIDGETS + ("ValidModel",)

SELECTION_SINGLE_CONTROL_WIDGETS = (
    "DropdownModel",
    "RadioButtonsModel",
    "SelectModel",
    "SelectionSliderModel",
    "ToggleButtonsModel",
)
SELECTION_MULTIPLE_CONTROL_WIDGETS = (
    "SelectionRangeSliderModel",
    "SelectMultipleModel",
)
SELECTION_CONTROL_WIDGETS = (
    SELECTION_SINGLE_CONTROL_WIDGETS + SELECTION_MULTIPLE_CONTROL_WIDGETS
)
SELECTION_OUTPUT_WIDGETS = SELECTION_CONTROL_WIDGETS

STRING_CONTROL_WIDGETS = (
    # "Text",  # open ended widgets not supported
    # "Textarea",  # open ended widgets not supported
)
STRING_OUTPUT_WIDGETS = (
    "LabelModel",
    # "HTMLModel",  # TODO
    # "HTMLMathModel",  # TODO
    # "ImageModel",  # TODO
)

OTHER_CONTROL_WIDGETS = (
    # "ButtonModel",  # TODO
    # "PlayModel",  # TODO
    # "DatePickerModel",  # TODO
    # "ColorPickerModel"  # TODO
)

CONTROL_WIDGETS = (
    NUMERIC_CONTROL_WIDGETS
    + BOOLEAN_CONTROL_WIDGETS
    + SELECTION_CONTROL_WIDGETS
    + STRING_CONTROL_WIDGETS
    + OTHER_CONTROL_WIDGETS
)

OUTPUT_WIDGETS = (
    NUMERIC_OUTPUT_WIDGETS
    + BOOLEAN_OUTPUT_WIDGETS
    + SELECTION_OUTPUT_WIDGETS
    + STRING_OUTPUT_WIDGETS
    + ("OutputModel",)
)

VALUE_WIDGETS = CONTROL_WIDGETS + OUTPUT_WIDGETS
LAYOUT_WIDGETS = (
    "BoxModel",
    "HBoxModel",
    "VBoxModel",
    "AccordionModel",
    "TabModel",
)



def test_target_widgets(preprocessor):
    preprocessor.exec_code("w = W.IntSlider(min=0, max=3)")
    preprocessor.exec_code("lbl = W.Label()")

    value_w, control_w = preprocessor.get_target_widgets()
    assert len(value_w) == 2
    assert len(control_w) == 1



import pytest


@pytest.fixture
def preprocessor():
    """
    A fixture to create an Illusionist.Preprocesor with an empty notebook
    """
    import nbformat

    from illusionist import preprocessor

    nb = nbformat.v4.new_notebook()
    p = preprocessor.IllusionistPreprocessor(nb=nb)

    # this_dir = os.path.abspath(os.path.dirname(__file__))
    # nb_file = os.path.join(this_dir, "widget-gallery.ipynb")

    # cell = nbformat.v4.new_code_cell("import ipywidgets")
    # nb.cells.append(cell)
    # nb = nbformat.read(nb_file, as_version=4)
    # print(nb)
    # p = preprocessor.IllusionistPreprocessor()

    p.execute(cleanup_kc=False)

    from illusionist import widgets
    from illusionist.utils import get_source

    p.exec_code(get_source(widgets))

    yield p

    p._cleanup_kernel()



from nbconvert.exporters.notebook import NotebookExporter

from illusionist.preprocessor import IllusionistPreprocessor


class IllusionistNotebookExporter(NotebookExporter):
    # Name for the menu item under "File -> Download as" in the IDE
    export_from_notebook = "Illusionist Notebook"
    preprocessors = [IllusionistPreprocessor]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['illusionist', 'illusionist.tests']

package_data = \
{'': ['*'],
 'illusionist': ['templates/illusionist/*', 'templates/illusionist/assets/*']}

install_requires = \
['ipywidgets>=7.6.5,<8.0.0',
 'nbclient>=0.5.0',
 'nbconvert>=6.2.0,<7',
 'nbformat>=5.0.7',
 'structlog>=21.1.0,<22.0.0']

entry_points = \
{'nbconvert.exporters': ['illusionist = illusionist:IllusionistHTMLExporter',
                         'illusionist-nb = '
                         'illusionist:IllusionistNotebookExporter']}

setup_kwargs = {
    'name': 'illusionist',
    'version': '0.3.0',
    'description': 'Interactive Jupyter Widgets without a Kernel',
    'long_description': '# illusionist\n\n[![pypi](https://badge.fury.io/py/illusionist.svg)](https://pypi.org/project/illusionist/)\n[![build](https://github.com/danielfrg/illusionist/workflows/test/badge.svg)](https://github.com/danielfrg/illusionist/actions/workflows/test.yml)\n[![docs](https://github.com/danielfrg/illusionist/workflows/docs/badge.svg)](https://github.com/danielfrg/illusionist/actions/workflows/docs.yml)\n[![coverage](https://codecov.io/gh/danielfrg/illusionist/branch/master/graph/badge.svg)](https://codecov.io/gh/danielfrg/illusionist?branch=master)\n[![license](https://img.shields.io/:license-Apache%202-blue.svg)](https://github.com/danielfrg/illusionist/blob/master/LICENSE.txt)\n\nIllusionist takes a Jupyter Notebook with widgets and converts it to a\nan HTML report that maintains the interactivity of the widgets without a\nrunning Jupyter kernel.\n\nIt does this by making all computation upfront and serializing all the possible outputs.\nIt generates a self-contained asset that you can easily drop into a file server\nand have an interactive report that scales.\n\nThe main idea of Jupyter Notebooks and Jupyter widgets is to make data closer\nto the code and data scientists while maintaining interactivity, they do a great job at that.\nIllusionist maintains the same development workflow Jupyter users are used to by using\nstandard Jupyter tooling such as `ipywidgets` and `nbconvert`.\nNo need to import anything in your notebook to generate an interactive report using illusionist,\njust run one `nbconvert` command.\n\nThe generated assets are easy to deploy, scale and have a big longevity by\nremoving a lot of deployment requirements and dependencies.\n\nLearn more and see examples in [the docs](https://illusionist.danielfrg.com/).\n',
    'author': 'Daniel Rodriguez',
    'author_email': 'None',
    'maintainer': 'Daniel Rodriguez',
    'maintainer_email': 'None',
    'url': 'https://github.com/danielfrg/mkdocs-jupyter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4',
}


setup(**setup_kwargs)



# This module is executed on the Kernel that also executes the notebook

try:
    # This will fail when executed in the Kernel is just for linting purposes
    import ipywidgets as W

    from .widgets import (
        BOOLEAN_OUTPUT_WIDGETS,
        CONTROL_WIDGETS,
        NUMERIC_OUTPUT_WIDGETS,
        SELECTION_OUTPUT_WIDGETS,
        STRING_OUTPUT_WIDGETS,
        VALUE_WIDGETS,
    )
except:
    pass


def get_widgets_ids(widgets=None, kind=None):
    """
    Return all widgets that are used in the notebook
    Parameters
    ----------
        widgets (list): List of widgets to
    Returns
    -------
        dictionary of `{"model_id": widget_object}`
    """
    # The `Widgets` class has a registry of all the widgets that were used
    all_widgets = widgets if widgets else W.Widget.widgets

    if kind is None:
        return list(all_widgets.keys())

    if kind == "value":
        value_widgets = []
        for model_id, obj in all_widgets.items():
            if isinstance(obj, VALUE_WIDGETS):
                value_widgets.append(model_id)
        return value_widgets

    if kind == "control":
        control_widgets = []
        for model_id, obj in all_widgets.items():
            if isinstance(obj, CONTROL_WIDGETS) and not obj.disabled:
                control_widgets.append(model_id)
        return control_widgets


def set_widget_value(widget_id, value):
    """
    Set the value of a widget
    """
    widget = W.Widget.widgets[widget_id]
    if isinstance(widget, NUMERIC_OUTPUT_WIDGETS):
        widget.value = value
        # raise Exception(widget, value, widget.value)
    elif isinstance(widget, BOOLEAN_OUTPUT_WIDGETS):
        widget.value = value
    elif isinstance(widget, SELECTION_OUTPUT_WIDGETS):
        widget.index = value
    elif isinstance(widget, STRING_OUTPUT_WIDGETS):
        return widget.value
    else:
        widget_type = type(widget)
        raise Exception(f"Cannot set value on widget type: 'f{widget_type}'")



import ipywidgets as W
import pytest

import illusionist.widgets as IW
from illusionist.preprocessor import possible_values


@pytest.mark.parametrize("class_", IW.NUMERIC_SINGLEVALUE_CONTROL_WIDGETS)
def test_possible_values_num_singlevalue(class_):
    w = class_(min=1, max=5)
    w_state = w.get_state()
    assert possible_values(w_state) == [1, 2, 3, 4, 5]

    w = class_(min=-3, max=2)
    w_state = w.get_state()
    assert possible_values(w_state) == [-3, -2, -1, 0, 1, 2]


@pytest.mark.parametrize("class_", IW.NUMERIC_MULTIVALUE_CONTROL_WIDGETS)
def test_possible_values_num_multivalue(class_):
    w = class_(min=0, max=3)
    w_state = w.get_state()
    a = [
        [0, 0],
        [0, 1],
        [0, 2],
        [0, 3],
        [1, 1],
        [1, 2],
        [1, 3],
        [2, 2],
        [2, 3],
        [3, 3],
    ]
    assert possible_values(w_state) == a


@pytest.mark.parametrize("class_", IW.BOOLEAN_CONTROL_WIDGETS)
def test_possible_values_bool(class_):
    w = class_()
    w_state = w.get_state()
    assert possible_values(w_state) == [True, False]


@pytest.mark.parametrize("class_", IW.SELECTION_SINGLEVALUE_CONTROL_WIDGETS)
def test_possible_values_selection_singlevalue(class_):
    options = ["a", "b", "c"]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [0, 1, 2]

    options = [("One", 1), ("Two", 2), ("Three", 3)]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [0, 1, 2]


@pytest.mark.parametrize("class_", [W.SelectionRangeSlider])
def test_possible_values_selection_multivalue_1(class_):
    options = ["a", "b", "c"]
    w = class_(options=options)
    w_state = w.get_state()
    a = [[0, 0], [0, 1], [0, 2], [1, 1], [1, 2], [2, 2]]
    assert possible_values(w_state) == a


@pytest.mark.parametrize("class_", [W.SelectMultiple])
def test_possible_values_selection_multivalue_2(class_):
    options = ["a"]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [(), (0,)]

    options = ["a", "b"]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [(), (0,), (1,), (0, 1)]

    options = ["a", "b", "c"]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [
        (),
        (0,),
        (1,),
        (2,),
        (0, 1),
        (0, 2),
        (1, 2),
        (0, 1, 2),
    ]



def test_affected_no_links(preprocessor):
    preprocessor.exec_code(
        """
w = W.IntSlider(min=0, max=3)
lbl = W.Label()
"""
    )

    _ = preprocessor.exec_code("w.model_id")
    w_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("lbl.model_id")
    lbl_id = preprocessor.eval_cell(_)

    value_w, control_w = preprocessor.get_target_widgets()
    links = preprocessor.get_affected_widgets(control_w, value_w)
    assert len(links) == 2
    assert w_id in links
    assert lbl_id in links
    for w_id, aff_by in links.items():
        assert len(aff_by) == 0


def test_affected_1_in_1_out(preprocessor):
    preprocessor.exec_code(
        """
slider = W.IntSlider(min=0, max=3)
lbl = W.Label()

def update(args):
    lbl.value = str(slider.value)

update(None)
slider.observe(update, "value")
"""
    )

    _ = preprocessor.exec_code("slider.model_id")
    slider_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("lbl.model_id")
    lbl_id = preprocessor.eval_cell(_)

    value_w, control_w = preprocessor.get_target_widgets()
    links = preprocessor.get_affected_widgets(control_w, value_w)
    assert len(links) == 2
    assert slider_id in links
    assert len(links[slider_id]) == 0
    assert len(links[lbl_id]) == 1


def test_affected_1_in_multi_out(preprocessor):
    preprocessor.exec_code(
        """
slider = W.IntSlider(min=0, max=3)
lbl = W.Label()
int = W.BoundedIntText(min=0, max=3)

def update(args):
    lbl.value = str(slider.value)
    int.value = slider.value

update(None)
slider.observe(update, "value")
"""
    )

    _ = preprocessor.exec_code("slider.model_id")
    slider_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("lbl.model_id")
    lbl_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("int.model_id")
    int_id = preprocessor.eval_cell(_)

    value_w, control_w = preprocessor.get_target_widgets()
    links = preprocessor.get_affected_widgets(control_w, value_w)
    assert len(links) == 3
    assert slider_id in links
    assert len(links[slider_id]) == 0
    assert len(links[lbl_id]) == 1
    assert len(links[int_id]) == 1


def test_affected_1_out_multi_in(preprocessor):
    preprocessor.exec_code(
        """
slider = W.IntSlider(min=0, max=3)
int = W.BoundedIntText(min=0, max=3)
lbl = W.Label()

def update(args):
    lbl.value = str(slider.value + int.value)

update(None)
slider.observe(update, "value")
int.observe(update, "value")
"""
    )

    _ = preprocessor.exec_code("slider.model_id")
    slider_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("lbl.model_id")
    lbl_id = preprocessor.eval_cell(_)
    _ = preprocessor.exec_code("int.model_id")
    int_id = preprocessor.eval_cell(_)

    value_w, control_w = preprocessor.get_target_widgets()
    links = preprocessor.get_affected_widgets(control_w, value_w)
    assert len(links) == 3
    assert slider_id in links
    assert len(links[slider_id]) == 0
    assert len(links[int_id]) == 0
    assert len(links[lbl_id]) == 2



import nbformat
from nbclient import NotebookClient
from nbclient.util import ensure_async, run_sync
from traitlets.config.application import Application

from illusionist.config import settings


class IllusionistClient(NotebookClient, Application):
    """
    Extends NBClient to add some utilities to run commands in the kernel
    """

    store_widget_state = True

    async def async_execute(self, reset_kc=False, **kwargs):
        """
        We overwrite and copy the original code for this function from
        nbclient.NotebookClient and change a couple of things

        We need to do this because output widgets are
        not being updated if I run code in a function after even
        after keeping the kernel alive (not sure why).
        """
        if reset_kc and self.km:
            await self._async_cleanup_kernel()
        self.reset_execution_trackers()

        async with self.async_setup_kernel(**kwargs):
            self.log.info(
                "Executing notebook with kernel: %s" % self.kernel_name
            )
            for index, cell in enumerate(self.nb.cells):
                # Ignore `'execution_count' in content` as it's always 1
                # when store_history is False
                await self.async_execute_cell(
                    cell, index, execution_count=self.code_cells_executed + 1
                )

            # ADDED
            await ensure_async(self.after_notebook())
            # NED

            msg_id = await ensure_async(self.kc.kernel_info())
            info_msg = await self.async_wait_for_reply(msg_id)
            self.nb.metadata["language_info"] = info_msg["content"][
                "language_info"
            ]
            self.set_widgets_metadata()

        return self.nb

    execute = run_sync(async_execute)

    def after_notebook(self):
        """
        This is called after the regular notebook cells have been executed
        """

    async def async_exec_code(self, source, write_cell=settings.dev_mode):
        """
        Execute code in a Kernel

        Parameters
        ----------
            source (str): Execute this code
            write_cell (bool, default=None): Write a new cell to the notebook

        Returns
        -------
            NotebookNode
        """
        cell = nbformat.NotebookNode()
        cell.cell_type = "code"
        cell.execution_count = self.code_cells_executed + 1
        cell.metadata = {}
        cell.outputs = []
        cell.source = source

        self.nb["cells"].append(cell)
        cell_index = len(self.nb["cells"]) - 1
        cell = await self.async_execute_cell(
            cell, cell_index, execution_count=cell.execution_count
        )

        if write_cell == False:
            # Delete created cell
            del self.nb["cells"][cell_index]

        return cell

    exec_code = run_sync(async_exec_code)

    def eval_cell(self, cell):
        """
        Run Python `eval` on a NotebookCell

        Parameters
        ----------
            cell (NotebookNode):

        Returns
        -------
            Python objects
        """
        output_txt = cell["outputs"][0]["data"]["text/plain"]
        return eval(output_txt)
